import 'dart:convert';
import 'dart:io';

import 'package:dart_amqp/dart_amqp.dart';
import 'package:s3i_flutter/src/auth/authentication_manager.dart';
import 'package:s3i_flutter/src/auth/tokens.dart';
import 'package:s3i_flutter/src/broker/broker_interfaces.dart';
import 'package:s3i_flutter/src/broker/messages/message.dart';
import 'package:s3i_flutter/src/exceptions/network_authentication_exception.dart';
import 'package:s3i_flutter/src/exceptions/s3i_exception.dart';

///Creates a new [BrokerAmqpConnector].
///
/// The instance is created with the [authManager] and the optional constructor
/// arguments in the [args] (the string-keys should match the name of the
/// parameters).
///
/// This is needed to use the same interface on web and other platforms.
///
/// The [args] have defaults for the normal message communication in the S3I:
/// - brokerHost = 'rabbitmq.s3i.vswf.dev',
/// - port = 5671,
/// - virtualHost = 's3i',
/// - maxConnectionAttempts = 3,
/// - reconnectWaitTime = const Duration(milliseconds: 1500),
/// - prefetchCount = -1,
/// - exchangeName = 'demo.direct',
/// - exchangeType = ExchangeType.DIRECT
/// onBadCertificate returns always true
ActiveBrokerInterface getActiveBrokerDefaultConnector(
    AuthenticationManager authManager,
    {Map<String, dynamic> args = const <String, dynamic>{}}) {
  final String brokerHost =
      args['brokerHost'] as String? ?? 'rabbitmq.s3i.vswf.dev';
  final int port = args['port'] as int? ?? 5671;
  final String virtualHost = args['virtualHost'] as String? ?? 's3i';
  final int maxConnectionAttempts = args['maxConnectionAttempts'] as int? ?? 3;
  final Duration reconnectWaitTime = args['reconnectWaitTime'] as Duration? ??
      const Duration(milliseconds: 1500);
  final int prefetchCount = args['prefetchCount'] as int? ?? -1;
  final String exchangeName = args['exchangeName'] as String? ?? 'demo.direct';
  final ExchangeType exchangeType =
      args['exchangeType'] as ExchangeType? ?? ExchangeType.DIRECT;

  return BrokerAmqpConnector(authManager,
      brokerHost: brokerHost,
      port: port,
      virtualHost: virtualHost,
      maxConnectionAttempts: maxConnectionAttempts,
      reconnectWaitTime: reconnectWaitTime,
      prefetchCount: prefetchCount,
      exchangeName: exchangeName,
      exchangeType: exchangeType,
      onBadCertificate: (X509Certificate cert) => true);
}

/// Creates a new [BrokerAmqpConnector] configured to be used for the
/// Event System.
///
/// See [getActiveBrokerDefaultConnector] for more information.
ActiveBrokerInterface getActiveBrokerEventConnector(
    AuthenticationManager authManager,
    {Map<String, dynamic> args = const <String, dynamic>{}}) {
  if (args.containsKey('exchangeName')) args['exchangeName'] = 'eventExchange';
  if (args.containsKey('exchangeType')) {
    args['exchangeType'] = ExchangeType.TOPIC;
  }
  return getActiveBrokerDefaultConnector(authManager, args: args);
}

/// This [ActiveBrokerInterface] implementation uses the native messaging
/// protocol of the S3I-Broker: AMQP.
///
/// For more information about AMQP see their
/// [specification](https://www.amqp.org/sites/amqp.org/files/amqp.pdf) or refer
/// to the documentation of the used
/// [dart_amqp package](https://pub.dev/packages/dart_amqp).
class BrokerAmqpConnector extends ActiveBrokerInterface {
  /// Creates a [BrokerAmqpConnector] which uses the [authManager] to receive
  /// tokens.
  BrokerAmqpConnector(AuthenticationManager authManager,
      {required this.brokerHost,
      required this.port,
      required this.virtualHost,
      required this.maxConnectionAttempts,
      required this.reconnectWaitTime,
      required this.prefetchCount,
      required this.exchangeName,
      required this.exchangeType,
      required this.onBadCertificate})
      : super(authManager);

  /// The host to connect to.
  final String brokerHost;

  /// The port of the amqp broker.
  final int port;

  /// The connection vhost that will be sent to the server.
  final String virtualHost;

  /// The max number of reconnection attempts.
  final int maxConnectionAttempts;

  /// The time between each reconnect attempt.
  final Duration reconnectWaitTime;

  /// The number of messages loaded to the client without ACK.
  ///
  /// If this number is < 0 no prefetch count is set => infinite prefetch.
  final int prefetchCount;

  /// The name of the amqp broker exchange.
  final String exchangeName;

  /// The type of the exchange, see
  /// [here](https://www.rabbitmq.com/tutorials/amqp-concepts.html#exchanges)
  /// for more information.
  final ExchangeType exchangeType;

  /// This callback is triggered if the certificate of the connection
  /// to the broker is invalid.
  final bool Function(X509Certificate)? onBadCertificate;

  /// Stores all endpoints and their consumer queues.
  final Map<String, Consumer> _endpointConsumer = <String, Consumer>{};

  Client? _amqpClient;

  Channel? _channel;

  Exchange? _exchange;

  bool _keepAlive = false;

  bool _inBrokerConstruction = false;

  /// Starts the connection to the S3I-Broker.
  ///
  /// Returns an empty string if everything is ready or the error occurred
  /// during the connection as string.
  Future<String> connectToBroker() async {
    if (_keepAlive || _endpointConsumer.isNotEmpty) return 'already connected';
    _keepAlive = true;
    _inBrokerConstruction = true;
    final String createInfo = await _establishConnectionToBroker()
        .then((_) => '')
        .catchError((Object e) => e.toString())
        .whenComplete(() {
      _inBrokerConstruction = false;
    });
    return createInfo;
  }

  /// Disconnects from the S3I-Broker.
  ///
  /// If the application doesn't need the broker, please consider disconnect
  /// from the broker. This cleans all consuming endpoints.
  Future<void> disconnectFromBroker() async {
    _keepAlive = false;
    _inBrokerConstruction = false;
    await _stopListeningToEndpoints(_endpointConsumer.keys);
    _endpointConsumer.clear();
    await _resetConnectionToBroker();
  }

  /// Reconnects to S3I-Broker
  ///
  /// Allows to reconnect to broker, e.g. after app resumes after long time
  Future<void> reconnectToBroker() async{
    _reestablishConnection();
  }

  ///Publishes the given [message] to all [endpoints].
  ///
  /// The endpoints should be in the correct format (s3ib://s3i:+ UUIDv4 for
  /// decrypted communication or s3ibs://s3i: + UUIDv4 for encrypted messages).
  ///
  /// Connects to the Broker if there is no connection open at the time.
  @override
  Future<void> sendMessage(Message message, Set<String> endpoints) async {
    if (_amqpClient == null || _channel == null || _exchange == null) {
      if (_inBrokerConstruction) {
        notifySendMessageFailed(message, S3IException('invalid broker state'));
        return;
      } else {
        final String creationInfo = await connectToBroker();
        if (creationInfo.isNotEmpty) {
          notifySendMessageFailed(message, S3IException(creationInfo));
          return;
        }
      }
    }
    for (final String edp in endpoints) {
      try {
        _exchange!.publish(jsonEncode(message.toJson()), edp);
        notifySendMessageSucceeded(message);
      } on ChannelException catch (e) {
        notifySendMessageFailed(message, e);
      }
    }
  }

  ///Starts consuming on the [endpoint].
  ///
  /// The endpoints should be in the correct format (s3ib://s3i:+ UUIDv4 for
  /// decrypted communication or s3ibs://s3i: + UUIDv4 for encrypted messages).
  ///
  /// Connects to the Broker if there is no connection open at the time.
  /// Allow to reconnect (e.g. additional Topic Subscribed). If enpoint present,
  /// stopListing and reconnect
  @override
  Future<void> startConsuming(String endpoint) async {
    if (_endpointConsumer.containsKey(endpoint)){
      _stopListeningToEndpoint(endpoint);
      _endpointConsumer.remove(endpoint);
    }
    if (_amqpClient == null || _channel == null || _exchange == null) {
      if (_inBrokerConstruction) {
        notifyConsumingFailed(endpoint, S3IException('invalid broker state'));
        return;
      } else {
        final String creationInfo = await connectToBroker();
        if (creationInfo.isNotEmpty) {
          notifyConsumingFailed(endpoint, S3IException(creationInfo));
          return;
        }
      }
    }
    _connectToEndpoint(endpoint).catchError((Object e) {
      notifyConsumingFailed(endpoint, S3IException(e.toString()));
    });
  }

  /// Stops consuming on the [endpoint].
  ///
  /// The endpoints should be in the correct format (s3ib://s3i:+ UUIDv4 for
  /// decrypted communication or s3ibs://s3i: + UUIDv4 for encrypted messages).
  ///
  /// Disconnects from the Broker if no one is listening.
  @override
  Future<void> stopConsuming(String endpoint) async {
    await _stopListeningToEndpoint(endpoint);
    _endpointConsumer.remove(endpoint);
    if (_endpointConsumer.isEmpty) disconnectFromBroker();
  }

  /// Creates a new connection to the broker.
  ///
  /// sets _client, _exchange and _channel
  ///
  /// Throws a [NetworkResponseException] if no token could be received. Throws
  /// [S3IException] if an error occurs during the setup of the amqp part.
  /// See dart_amqp for more exceptions that could be thrown.
  ///
  /// Make sure to set [_inBrokerConstruction] before and after.
  Future<void> _establishConnectionToBroker() async {
    AccessToken token;
    try {
      token = await authManager.getAccessToken();
      // trigger keep alive
      if (_keepAlive) _keepConnectionAlive(token);
    } on Exception catch (e) {
      throw NetworkAuthenticationException(e);
    }
    final ConnectionSettings setting = ConnectionSettings(
        host: brokerHost,
        port: port,
        virtualHost: virtualHost,
        authProvider: PlainAuthenticator(' ', token.originalToken),
        tlsContext: SecurityContext(),
        onBadCertificate: onBadCertificate,
        maxConnectionAttempts: maxConnectionAttempts,
        reconnectWaitTime: reconnectWaitTime);
    _amqpClient = Client(settings: setting);
    if (_amqpClient == null) throw S3IException('amqp client is null');
    try {
      _channel = await _amqpClient!.channel();
      if (prefetchCount >= 0) {
        _channel = await _channel!.qos(null, prefetchCount);
      }
      _exchange =
          await _channel!.exchange(exchangeName, exchangeType, passive: true);
    } on Exception catch (e) {
      throw S3IException('amqp package error: $e');
    }
  }

  /// Closes the connection and sets _client, _exchange and _channel to null.
  Future<void> _resetConnectionToBroker() async {
    if (_amqpClient != null) {
      await _amqpClient!.close();
    }
    _amqpClient = null;
    _channel = null;
    _exchange = null;
  }

  /// Should only be called when _client, _channel, _exchange != null.
  ///
  /// Connects to the endpoint and adds the consumer to the map.
  Future<void> _connectToEndpoint(String endpoint,
      {bool passive = true, bool durable = true}) async {
    if (_amqpClient == null || _channel == null || _exchange == null) return;
    final Queue receiveQueue =
        await _channel!.queue(endpoint, passive: passive, durable: durable);
    final Consumer consume = await receiveQueue.consume();
    consume.listen((AmqpMessage event) {
      newMessageReceived(event.payloadAsString);
    });
    _endpointConsumer[endpoint] = consume;
  }

  /// Cancels all consumers at [endpoints].
  Future<void> _stopListeningToEndpoints(Iterable<String> endpoints) async {
    for (final String end in endpoints) {
      await _stopListeningToEndpoint(end);
    }
  }

  /// Cancels the consumer at [endpoint].
  Future<void> _stopListeningToEndpoint(String endpoint) async {
    if (_endpointConsumer.containsKey(endpoint)) {
      final Consumer consumer = _endpointConsumer[endpoint]!;
      await consumer.cancel();
    }
  }

  /// Closes all connections/subscriptions to the broker and reestablish them
  /// afterwards. Distinguishes between DefaultConnector and EventConnector
  Future<void> _reestablishConnection() async {
    final Set<String> oldSubEndpoints = _endpointConsumer.keys.toSet();
    await _stopListeningToEndpoints(oldSubEndpoints);
    _endpointConsumer.clear();
    await _resetConnectionToBroker();
    //rebuild all old connections
    _inBrokerConstruction = true;
    await _establishConnectionToBroker().whenComplete(() {
      _inBrokerConstruction = false;
    });
    if(exchangeType == ExchangeType.TOPIC){
      recreateTopicBinding();
    }
    else{
      for (final String end in oldSubEndpoints) {
        await _connectToEndpoint(end).catchError((Object e) {
          notifyConsumingFailed(end, S3IException(e.toString()));
        });
      }
    }
  }

  /// Checks the expiration time of the token and reconnects to the Broker
  /// when it's expired.
  Future<void> _keepConnectionAlive(AccessToken token) async {
    Future<void>.delayed(token.timeTillExpiration()).then((dynamic _) {
      if (_keepAlive) _reestablishConnection();
    });
  }
}
